unit clWebUpdate;

interface

uses
  clDownloader, Classes, msxml, clDC, clDCUtils, clMultiDC;

type
  TclVersionFormat = (vfStandard, vfNumber);

  TclActualInfo = class
  private
    FSuccess: string;
    FUpdateDate: string;
    FVersion: string;
    FFileName: string;
  public
    property FileName: string read FFileName;
    property Version: string read FVersion;
    property Success: string read FSuccess;
    property UpdateDate: string read FUpdateDate;
  end;

  TclUpdateInfo = class
  private
    FURL: string;
    FProductURL: string;
    FRunParameters: string;
    FProductName: string;
    FVersion: string;
    FAuthor: string;
    FSize: string;
    FEmail: string;
  public
    property URL: string read FURL;
    property Size: string read FSize;
    property Version: string read FVersion;
    property RunParameters: string read FRunParameters;
    property ProductName: string read FProductName;
    property Author: string read FAuthor;
    property ProductURL: string read FProductURL;
    property Email: string read FEmail;
  end;


  TclHasUpdateEvent = procedure (Sender: TObject; const NewVersion, ActualVersion: string;
    var CanUpdate: Boolean) of object;
  TclDownloadEvent = procedure (Sender: TObject; const DownloadURL, DownloadFile: string) of object;
  TclRunUpdateEvent = procedure (Sender: TObject; var RunFile, RunParameters, RunOperation: string) of object;
  TclTerminatingEvent = procedure (Sender: TObject; var CanTerminate: Boolean) of object;
  TclDownloadProgressEvent = procedure (Sender: TObject; Downloaded, Total: Integer) of object;
  TclShowInfoEvent = procedure (Sender: TObject; ActualInfo: TclActualInfo; UpdateInfo: TclUpdateInfo;
    var CanUpdate, Handled: Boolean) of object;
  TclGetUpdateInfoEvent = procedure (Sender: TObject; ActualInfo: TclActualInfo;
    UpdateInfo: TclUpdateInfo) of object;

  TclWebUpdate = class(TComponent)
  private
    FDownloader: TclDownloader;
    FIsDownloading: Boolean;
    FUpdateURL: string;
    FLastUpdateInfoFile: string;
    FVersionFormat: TclVersionFormat;
    FUserName: string;
    FPassword: string;
    FNeedTerminate: Boolean;
    FRunUpdate: Boolean;
    FNeedShowInfo: Boolean;
    FActualInfo: TclActualInfo;
    FUpdateInfo: TclUpdateInfo;
    FOnDownloadProgress: TclDownloadProgressEvent;
    FOnError: TclOnError;
    FOnHasUpdate: TclHasUpdateEvent;
    FOnTerminating: TclTerminatingEvent;
    FOnRunUpdate: TclRunUpdateEvent;
    FOnAfterDownload: TclDownloadEvent;
    FOnBeforeDownload: TclDownloadEvent;
    FOnShowInfo: TclShowInfoEvent;
    FOnNoUpdatesFound: TNotifyEvent;
    FOnGetUpdateInfo: TclGetUpdateInfoEvent;
    procedure DoOnIsBusyChanged(Sender: TObject);
    procedure DoOnError(Sender: TObject; const Error: string; ErrorCode: Integer);
    procedure DoOnDataItemProceed(Sender: TObject; ResourceInfo: TclResourceInfo;
      AStateItem: TclResourceStateItem; CurrentData: PChar; CurrentDataSize: Integer);
    function GetIsBusy(): Boolean;
    function CheckUpdateVersion(const ANewVersion, ActualVersion: string): Boolean;
    function HasUpdate(): Boolean;
    procedure GetUpdateInfo();
    procedure StartDownloading();
    procedure StartUpdating();
    procedure StoreLastUpdateInfo(ASuccess: Boolean);
    function ShowInfo(): Boolean;
  protected
    procedure DoGetUpdateInfo();
    procedure DoHasUpdate(const NewVersion, ActualVersion: string; var CanUpdate: Boolean); dynamic;
    procedure DoBeforeDownload(const DownloadURL, DownloadFile: string); dynamic;
    procedure DoAfterDownload(const DownloadURL, DownloadFile: string); dynamic;
    procedure DoShowInfo(var CanUpdate, Handled: Boolean); dynamic;
    procedure DoRunUpdate(var RunFile, RunParameters, RunOperation: string); dynamic;
    procedure DoTerminating(var CanTerminate: Boolean); dynamic;
    procedure DoDownloadProgress(Downloaded, Total: Integer); dynamic;
    procedure DoError(const Error: string; ErrorCode: Integer); dynamic;
    procedure DoNoUpdatesFound();
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy(); override;
    procedure Start();
    procedure Stop();
    property IsBusy: Boolean read GetIsBusy;
    property ActualInfo: TclActualInfo read FActualInfo;
    property UpdateInfo: TclUpdateInfo read FUpdateInfo;
  published
    property UpdateURL: string read FUpdateURL write FUpdateURL;
    property LastUpdateInfoFile: string read FLastUpdateInfoFile write FLastUpdateInfoFile;
    property VersionFormat: TclVersionFormat read FVersionFormat write FVersionFormat default vfStandard;
    property UserName: string read FUserName write FUserName;
    property Password: string read FPassword write FPassword;
    property NeedTerminate: Boolean read FNeedTerminate write FNeedTerminate default False;
    property NeedShowInfo: Boolean read FNeedShowInfo write FNeedShowInfo default True;
    property RunUpdate: Boolean read FRunUpdate write FRunUpdate default False;
    property OnGetUpdateInfo: TclGetUpdateInfoEvent read FOnGetUpdateInfo write FOnGetUpdateInfo;
    property OnShowInfo: TclShowInfoEvent read FOnShowInfo write FOnShowInfo;
    property OnHasUpdate: TclHasUpdateEvent read FOnHasUpdate write FOnHasUpdate;
    property OnBeforeDownload: TclDownloadEvent read FOnBeforeDownload write FOnBeforeDownload;
    property OnAfterDownload: TclDownloadEvent read FOnAfterDownload write FOnAfterDownload;
    property OnRunUpdate: TclRunUpdateEvent read FOnRunUpdate write FOnRunUpdate;
    property OnTerminating: TclTerminatingEvent read FOnTerminating write FOnTerminating;
    property OnDownloadProgress: TclDownloadProgressEvent read FOnDownloadProgress write FOnDownloadProgress;
    property OnError: TclOnError read FOnError write FOnError;
    property OnNoUpdatesFound: TNotifyEvent read FOnNoUpdatesFound write FOnNoUpdatesFound;
  end;

procedure Register();

implementation

uses
  SysUtils, Math, ShellAPI, Forms, Windows, clUpdateInfoForm, clSingleDC;

procedure Register();
begin
  RegisterComponents('Clever Internet Suite', [TclWebUpdate]);
end;

{ TclWebUpdate }

constructor TclWebUpdate.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FActualInfo := TclActualInfo.Create();
  FUpdateInfo := TclUpdateInfo.Create();
  FDownloader := TclDownloader.Create(nil);
  FDownloader.LocalFolder := '.\';
  FDownloader.OnIsBusyChanged := DoOnIsBusyChanged;
  FDownloader.OnError := DoOnError;
  FDownloader.OnDataItemProceed := DoOnDataItemProceed;
  FLastUpdateInfoFile := 'lastupdate.xml';
  FVersionFormat := vfStandard;
  FNeedShowInfo := True;
end;

destructor TclWebUpdate.Destroy();
begin
  FDownloader.Free();
  FUpdateInfo.Free();
  FActualInfo.Free();
  inherited Destroy();
end;

procedure TclWebUpdate.DoOnIsBusyChanged(Sender: TObject);
var
  Success: Boolean;
begin
  if FDownloader.IsBusy then Exit;
  if FIsDownloading then
  begin
    Success := (FDownloader.ResourceState.LastStatus = psSuccess);
    StoreLastUpdateInfo(Success);
    if Success then
    begin
      DoAfterDownload(FDownloader.URL, FDownloader.LocalFile);
      FIsDownloading := False;
      StartUpdating();
    end;
  end else
  if (FDownloader.ResourceState.LastStatus = psSuccess) then
  begin
    FIsDownloading := True;
    GetUpdateInfo();
    if HasUpdate() then
    begin
      if ShowInfo() then
      begin
        StartDownloading();
      end;
    end else
    begin
      DoNoUpdatesFound();
    end;
  end;
end;

function TclWebUpdate.GetIsBusy: Boolean;
begin
  Result := FDownloader.IsBusy;
end;

procedure TclWebUpdate.Start();
begin
  FIsDownloading := False;
  FDownloader.URL := UpdateURL;
  FDownloader.UserName := UserName;
  FDownloader.Password := Password;
  FDownloader.Start();
end;

function TclWebUpdate.CheckUpdateVersion(const ANewVersion, ActualVersion: string): Boolean;
  function MakeCompleteNumber(S: string): Integer;
  var
    i, Cnt, Len, OldPos: Integer;
  begin
    S := Trim(S);
    Result := 0;
    Len := Length(S);
    Cnt := 0;
    OldPos := Len + 1;
    for i := Len downto 1 do
    begin
      if (S[i] = '.') then
      begin
        Result := Result + StrToIntDef('0' + system.Copy(S, i + 1, OldPos - i - 1), 0) * Round(Power(1000, Cnt));
        Inc(Cnt);
        OldPos := i;
      end else
      if (i = 1) then
      begin
        Result := Result + StrToIntDef('0' + system.Copy(S, i, OldPos - i), 0) * Round(Power(1000, Cnt));
      end;
    end;
  end;

begin
  if (VersionFormat = vfStandard) then
  begin
    Result := (MakeCompleteNumber(ANewVersion) > MakeCompleteNumber(ActualVersion));
  end else
  begin
    Result := StrToIntDef(ANewVersion, 0) > StrToIntDef(ActualVersion, 0);
  end;
end;

function TclWebUpdate.HasUpdate(): Boolean;
begin
  Result := True;
  if (ActualInfo.Version <> '') then
  begin
    if (ActualInfo.Success = 'yes') then
    begin
      Result := CheckUpdateVersion(UpdateInfo.Version, ActualInfo.Version);
    end;
  end;
  DoHasUpdate(UpdateInfo.Version, ActualInfo.Version, Result);
end;

procedure TclWebUpdate.GetUpdateInfo();
var
  Dom: IXMLDomDocument;
begin
  Dom := CoDOMDocument.Create();
  Dom.load(FDownloader.LocalFile);
  UpdateInfo.FURL := Dom.selectSingleNode('updateinfo/update/@url').text;
  UpdateInfo.FRunParameters := Dom.selectSingleNode('updateinfo/update/@runparameters').text;
  UpdateInfo.FVersion := Dom.selectSingleNode('updateinfo/update/@version').text;
  UpdateInfo.FSize := Dom.selectSingleNode('updateinfo/update/@size').text;
  UpdateInfo.FProductURL := Dom.selectSingleNode('updateinfo/product/@url').text;
  UpdateInfo.FProductName := Dom.selectSingleNode('updateinfo/product/@name').text;
  UpdateInfo.FAuthor := Dom.selectSingleNode('updateinfo/product/@author').text;
  UpdateInfo.FEmail := Dom.selectSingleNode('updateinfo/product/@email').text;
  ActualInfo.FSuccess := '';
  ActualInfo.FUpdateDate := '';
  ActualInfo.FVersion := '';
  ActualInfo.FFileName := '';
  if FileExists(LastUpdateInfoFile) then
  begin
    Dom.load(LastUpdateInfoFile);
    ActualInfo.FSuccess := Dom.selectSingleNode('update/@success').text;
    ActualInfo.FUpdateDate := Dom.selectSingleNode('update/@updatedate').text;
    ActualInfo.FVersion := Dom.selectSingleNode('update/@version').text;
    ActualInfo.FFileName := Dom.selectSingleNode('update/@file').text;
  end;
  DoGetUpdateInfo();
end;

function TclWebUpdate.ShowInfo(): Boolean;
var
  Handled: Boolean;
begin
  Result := True;
  if not NeedShowInfo then Exit;
  Handled := False;
  DoShowInfo(Result, Handled);
  if not Handled then
  begin
    Result := TfrmUpdateInfo.ShowInfo(Self);
  end;
end;

procedure TclWebUpdate.StartDownloading();
begin
  FDownloader.URL := UpdateInfo.URL;
  FDownloader.UserName := UserName;
  FDownloader.Password := Password;
  DoBeforeDownload(FDownloader.URL, FDownloader.LocalFile);
  FDownloader.Start();
end;

procedure TclWebUpdate.StartUpdating();
var
  RunFile, RunParams, RunOperation: string;
  CanTerminate: Boolean;
  res: Integer;
begin
  if RunUpdate then
  begin
    RunFile := FDownloader.LocalFile;
    RunParams := UpdateInfo.RunParameters;
    RunOperation := 'open';
    DoRunUpdate(RunFile, RunParams, RunOperation);
    res := ShellExecute(Application.Handle, PChar(RunOperation),
      PChar(RunFile), PChar(RunParams), nil, SW_SHOWNORMAL);
    if (res <= 32) then
    begin
      DoError(Format('The executing of %s file was returned with errors (%d). ', [RunFile, res]), res);
    end;
  end;
  if NeedTerminate then
  begin
    CanTerminate := True;
    DoTerminating(CanTerminate);
    if CanTerminate then
    begin
      Application.Terminate();
    end;
  end;
end;

procedure TclWebUpdate.StoreLastUpdateInfo(ASuccess: Boolean);
const
  BooleanMap: array[Boolean] of string = ('no', 'yes');
var
  Dom: IXMLDomDocument;
  Node: IXMLDomElement;
begin
  ActualInfo.FFileName := FDownloader.LocalFile;
  ActualInfo.FVersion := UpdateInfo.Version;
  ActualInfo.FUpdateDate := FormatDateTime('d mmm yyyy', Date());
  ActualInfo.FSuccess := BooleanMap[ASuccess];
  if (LastUpdateInfoFile <> '') then
  begin
    Dom := CoDOMDocument.Create();
    Node := Dom.createElement('update');
    Dom.appendChild(Node);
    Node.setAttribute('file', ActualInfo.FileName);
    Node.setAttribute('version', ActualInfo.Version);
    Node.setAttribute('success', ActualInfo.Success);
    Node.setAttribute('updatedate', ActualInfo.UpdateDate);
    Dom.save(LastUpdateInfoFile);
  end;
end;

procedure TclWebUpdate.Stop();
begin
  FDownloader.Stop();
end;

procedure TclWebUpdate.DoOnDataItemProceed(Sender: TObject;
  ResourceInfo: TclResourceInfo; AStateItem: TclResourceStateItem; CurrentData: PChar;
  CurrentDataSize: Integer);
begin
  if FIsDownloading then
  begin
    DoDownloadProgress(AStateItem.ResourceState.BytesProceed, ResourceInfo.Size);
  end;
end;

procedure TclWebUpdate.DoOnError(Sender: TObject; const Error: string; ErrorCode: Integer);
begin
  DoError(Error, ErrorCode);
end;

procedure TclWebUpdate.DoAfterDownload(const DownloadURL, DownloadFile: string);
begin
  if Assigned(OnAfterDownload) then
  begin
    OnAfterDownload(Self, DownloadURL, DownloadFile);
  end;
end;

procedure TclWebUpdate.DoBeforeDownload(const DownloadURL, DownloadFile: string);
begin
  if Assigned(OnBeforeDownload) then
  begin
    OnBeforeDownload(Self, DownloadURL, DownloadFile);
  end;
end;

procedure TclWebUpdate.DoDownloadProgress(Downloaded, Total: Integer);
begin
  if Assigned(OnDownloadProgress) then
  begin
    OnDownloadProgress(Self, Downloaded, Total);
  end;
end;

procedure TclWebUpdate.DoError(const Error: string; ErrorCode: Integer);
begin
  if Assigned(OnError) then
  begin
    OnError(Self, Error, ErrorCode);
  end;
end;

procedure TclWebUpdate.DoHasUpdate(const NewVersion, ActualVersion: string; var CanUpdate: Boolean);
begin
  if Assigned(OnHasUpdate) then
  begin
    OnHasUpdate(Self, NewVersion, ActualVersion, CanUpdate);
  end;
end;

procedure TclWebUpdate.DoRunUpdate(var RunFile, RunParameters, RunOperation: string);
begin
  if Assigned(OnRunUpdate) then
  begin
    OnRunUpdate(Self, RunFile, RunParameters, RunOperation);
  end;
end;

procedure TclWebUpdate.DoShowInfo(var CanUpdate, Handled: Boolean);
begin
  if Assigned(OnShowInfo) then
  begin
    OnShowInfo(Self, ActualInfo, UpdateInfo, CanUpdate, Handled);
  end;
end;

procedure TclWebUpdate.DoTerminating(var CanTerminate: Boolean);
begin
  if Assigned(OnTerminating) then
  begin
    OnTerminating(Self, CanTerminate);
  end;
end;

procedure TclWebUpdate.DoNoUpdatesFound;
begin
  if Assigned(OnNoUpdatesFound) then
  begin
    OnNoUpdatesFound(Self);
  end;
end;

procedure TclWebUpdate.DoGetUpdateInfo;
begin
  if Assigned(OnGetUpdateInfo) then
  begin
    OnGetUpdateInfo(Self, ActualInfo, UpdateInfo);
  end;
end;

end.
